// Digital Cash
// Steven Kappes / Justin Fiedler

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import java.io.*;
import java.util.*;
import java.math.*;
import java.net.*;
import javax.net.ssl.*;
import java.security.cert.*;
import java.security.cert.Certificate;
import java.security.*;
import javax.crypto.*;
import org.w3c.dom.*;

public class CustomerGUI implements ActionListener {
	static final public  String gamsdir= "./"; 
	static final public String fileName = "customer.txt";

	static private org.apache.xerces.dom.DOMImplementationImpl DOM = new org.apache.xerces.dom.DOMImplementationImpl();
	static private DocumentType docType = DOM.createDocumentType("message", "http://localhost/message.dtd", "message.dtd");

	static private Receiver rr;
	static private XMLSocketThread xmlSocket;

	final int MERCH_PORT = 3347;

	static Customer me;

	// Components
	String		amount, balance, merchant, mo;
	JLabel		amountLabel, balanceLabel, moLabel, merchantLabel;
	JLabel emptyLabel = new JLabel(" ");
	JButton		withdrawButton, getBalanceButton, sendToMerchantButton;
	JTextField	amountTF, balanceTF, merchantTF;
	JComboBox		moList;

	ArrayList<String> moNames = new ArrayList<String>();

	static class MyKeyManager implements X509KeyManager {
		X509KeyManager x509mgr;
		String alias;

		MyKeyManager(X509KeyManager x509mgr, String alias) {
			this.x509mgr = x509mgr;
			this.alias = alias;
			String[] aliases = x509mgr.getClientAliases("RSA", null);
			for (int i=0; i<aliases.length; ++i) {
				//System.out.println("found alias: " + aliases[i]);
			}
		}
		// force desired alias
		public String chooseClientAlias(String[] keytype, Principal[] issuers, Socket socket) {
			return alias;
		}
		public String chooseServerAlias(String keyType, Principal[] issuers, Socket socket) {
			return alias;
			//return x509mgr.chooseServerAlias(keyType, issuers, socket) 
		}
		public X509Certificate[] getCertificateChain(String alias) {
			return x509mgr.getCertificateChain(alias);
		}
		public PrivateKey getPrivateKey(String alias) {
			return x509mgr.getPrivateKey(alias);
		}
		public String[] getClientAliases(String keyType, Principal[] issuers) {
			return new String[] {alias};
		}
		public String[] getServerAliases(String keyType, Principal[] issuers) {
			return new String[] {alias};
		}
	}

	final static String LOOKANDFEEL = null;

	public Component createComponents() {
		//CREATE COMPONENTS
		// Amount
		amountLabel = new JLabel("Amount");
	    	amountTF = new JTextField(20);
	    	amountTF.addActionListener(this);
		
		// Withdraw
		withdrawButton = new JButton("Withdraw");
		withdrawButton.setActionCommand("withdraw");
		withdrawButton.addActionListener(this);

		// Balance
		balanceLabel = new JLabel("Balance");
		balanceTF = new JTextField(20);
		balanceTF.setEditable(false);
		balanceTF.addActionListener(this);

		// Get Balance
		getBalanceButton = new JButton("Get Balance");
		getBalanceButton.setActionCommand("getbalance");
		getBalanceButton.addActionListener(this);
		
		// MO List
		moLabel = new JLabel("Money Orders");	
		String[] entries = {"none"};
		moList = new JComboBox(entries);
		moList.setSelectedIndex(0);
		moList.addActionListener(this);

		// Merchant
		//merchantLabel = new JLabel("Merchant");
	    	//merchantTF = new JTextField(20);
	    	//merchantTF.addActionListener(this);
		
		// Send MO To Merchant
		sendToMerchantButton = new JButton("Send");
		sendToMerchantButton.setActionCommand("sendtomerchant");
		sendToMerchantButton.addActionListener(this);

		//ADD TO PANE
		JPanel pane = new JPanel(new GridLayout(0, 3));
		pane.add(amountLabel);
		pane.add(amountTF);
		pane.add(withdrawButton);
		
		pane.add(balanceLabel);
		pane.add(balanceTF);
		pane.add(getBalanceButton);

		pane.add(moLabel);
		pane.add(moList);
		//pane.add(emptyLabel);

		//pane.add(merchantLabel);
		//pane.add(merchantTF);
		pane.add(sendToMerchantButton);

		pane.setBorder(BorderFactory.createEmptyBorder( 10, 10, 10, 10)); //(top, left, bottom, right)
		return pane;	
	}
	
	public void actionPerformed(ActionEvent e) {
		if("withdraw".equals(e.getActionCommand())) { //WITHDRAW MONEY
			this.requestMoney();
		} else if("getbalance".equals(e.getActionCommand())) { //GET BALANCE
			this.requestBalance();
		} else if("sendtomerchant".equals(e.getActionCommand())) { //SEND TO MERCHANT
			this.sendToMerchant();
		}
	}

	public static void setComObjects(Receiver r, XMLSocketThread xmlSoc)
	{
		rr = r;
		xmlSocket = xmlSoc;
	}

	public static void createAndShowGUI() {
		//Set the look and feel.
		try{
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch(Exception e){ }

		//Make sure we have nice window decorations.
		JFrame.setDefaultLookAndFeelDecorated(true);

		//Create and set up the window.
		JFrame frame = new JFrame("Customer GUI");
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

		CustomerGUI app = new CustomerGUI();
		Component contents = app.createComponents();
		frame.getContentPane().add(contents, BorderLayout.CENTER);

		//Display the window.
		frame.pack();
		frame.setVisible(true);
	}

	public CustomerGUI()
	{
		me = new Customer();
	}

	public static void run() {
		//Schedule a job for the event-dispatching thread:
		//creating and showing this application's GUI.
		javax.swing.SwingUtilities.invokeLater(new Runnable() { public void run() { createAndShowGUI(); } });
	}

	public void requestBalance()
	{
		try {
			String customerID = DigitalCash.getMyIdentity().toString();
			String password = DigitalCash.MyPassword().toString();

			// Create the document
			Document doc = XMLHelper.sendBalance(customerID, password, -1);
			xmlSocket.send(doc);

			// Get the response
			doc = rr.get();
			doc.normalize();
			XMLHelper.stripText(doc);

			if (!XMLHelper.isValidMessage(doc, 0)) return;

			String amt = XMLHelper.getItem(doc, "balance", "0");
			if (amt != null)
				balanceTF.setText("$" + amt);
		} catch (Exception e) {
			System.err.println("Unable to get balance");
			System.exit(1);
		}
	}

	public void requestMoney()
	{
		try {
			String customerID = DigitalCash.getMyIdentity().toString();
			String password = DigitalCash.MyPassword().toString();
			Document doc;

			// Get the requested amount
			int amt;
			String amtStr = amountTF.getText();
			if ((amtStr.length() == 0) || (amtStr.charAt(0) < '0') || (amtStr.charAt(0) > '9')) {
				System.out.println("Invalid amount to withdraw");
				return;
			}

			amt = Integer.parseInt(amtStr);
			if (amt <= 0) {
				System.out.println("Invalid amount to withdraw");
				return;
			}

			// Send money orders to the bank
			ArrayList<ByteArray> encryptedMOs = me.generateEncryptedMoneyOrders(amt);
			ArrayList<ByteArray> blindHashes = me.getBlindedHashes();

			doc = XMLHelper.sendEncryptedOrders(customerID, password, encryptedMOs, blindHashes);
			xmlSocket.send(doc);

			// Get the index of the money order to be signed
			int moIndex;
			doc = rr.get();
			doc.normalize();
			XMLHelper.stripText(doc);

			if (!XMLHelper.isValidMessage(doc, 2)) {
				me.setHasRequest(false);
				return;
			}

			moIndex = Integer.parseInt(XMLHelper.getItem(doc, "MOIndex", "0"));
			me.setFutureOrderIndex(moIndex);

			// Send the keys and blinding factors
			ArrayList<SecretKey> keys = me.getAllEncryptionKeys(moIndex);
			ArrayList<BigInteger> blinds = me.getAllBlinds(moIndex);
			ArrayList<ArrayList<SecretKey>> identKeysLeft = new ArrayList<ArrayList<SecretKey>>();
			ArrayList<ArrayList<SecretKey>> identKeysRight = new ArrayList<ArrayList<SecretKey>>();

			for (int i = 0; i < DigitalCash.N; i++) {
				if (i != moIndex) {
					identKeysLeft.add(me.getIdentityKeys(i, true));
					identKeysRight.add(me.getIdentityKeys(i, false));
				}
			}
			doc = XMLHelper.sendKeysAndBlinds(customerID, password, keys, blinds, identKeysLeft, identKeysRight);
			xmlSocket.send(doc);

			// Get the signed hash
			ByteArray signedHash;
			doc = rr.get();
			doc.normalize();
			XMLHelper.stripText(doc);

			if (!XMLHelper.isValidMessage(doc, 4)) {
				me.setHasRequest(false);
				return;
			}

			signedHash = new ByteArray(XMLHelper.getItem(doc, "SignedHash", "0"));
			if (!me.receivedSignedOrder(signedHash)) {
				System.err.println("Money order request failed");
				me.setHasRequest(false);
				return;
			}
			updateMoneyOrderList();
		} catch (Exception e) {
			System.err.println("CustomerGUI.requestMoney: Unable to get digital cash");
			e.printStackTrace();
			System.exit(1);
		}
	}

	public void sendToMerchant()
	{
		int moIndex = moList.getSelectedIndex();
		if ((moIndex >= 0) && (moIndex < me.validMoneyOrders().size())) {
			try {
				SSLContext sslc = SSLContext.getInstance("SSL");
				this.initKeyStore(sslc);
				SSLSocket sock = (SSLSocket)sslc.getSocketFactory().createSocket(InetAddress.getByName("localhost"), MERCH_PORT);

				sock.setEnabledCipherSuites(sock.getSupportedCipherSuites());
				sock.setNeedClientAuth(true);
				SSLSession sess = sock.getSession();

				try {
					Certificate[] certs = sess.getPeerCertificates();
					System.out.println("Merchant authenticated");
				} catch (SSLPeerUnverifiedException ex) {
					System.out.println("Merchant NOT authenticated");
					return;
				}

				XMLSocketThread xmlSock = new XMLSocketThread(sock);
				Receiver r2 = new Receiver(xmlSock);
				xmlSock.setDOMProcessor(r2);
				xmlSock.start();

				Document doc;
	
				// Send the money order
				MoneyOrder mo = me.validMoneyOrders().get(moIndex);
				doc = XMLHelper.sendMoneyAndSignedHash(mo.intString(), mo.getSignedHash());
				xmlSock.send(doc);

				// Get the response
				doc = r2.get();
				doc.normalize();
				XMLHelper.stripText(doc);

				if (!XMLHelper.isValidMessage(doc, 6)) return;

				ByteArray moStr = new ByteArray(XMLHelper.getItem(doc, "MoneyOrderUniqueStr", "0"));
				ArrayList<Integer> unlockSecs = new ArrayList<Integer>();
				for (int i = 0; i < DigitalCash.N; i++) {
					unlockSecs.add(new Integer(XMLHelper.getItem(doc, "unlocksecret", String.valueOf(i))));
				}

				ArrayList<SecretKey> identKeys = me.getIdentKeys(moStr, unlockSecs);
				doc = XMLHelper.sendIdentityKeys(identKeys);
				xmlSock.send(doc);

				// Wait to see if the money was accepted
				doc = r2.get();
				doc.normalize();
				XMLHelper.stripText(doc);
				XMLHelper.isValidMessage(doc, -1); // Print the message
			} catch (Exception e) {
				System.err.println("Unable to send money order: " + e.getMessage());
				e.printStackTrace();
				return;
			}
		} else {
			System.err.println("Select a money order");
		}
	}

	void initKeyStore(SSLContext sslc)
	{
		try {
			KeyStore ks = KeyStore.getInstance("JKS");
			ks.load(new FileInputStream("customerstore"), "foofoo".toCharArray());
			KeyManagerFactory kmf = KeyManagerFactory.getInstance("SunX509");
			kmf.init(ks, "foofoo".toCharArray());

			KeyStore ks2 = KeyStore.getInstance("JKS");
			ks2.load(new FileInputStream("customertrust"), "barbar".toCharArray());
			TrustManagerFactory tmf = TrustManagerFactory.getInstance("SunX509");
			tmf.init(ks2);

			KeyManager[] kms = kmf.getKeyManagers();
			for (int i=0; i<kms.length; ++i) {
				if (kms[i] instanceof X509KeyManager) {
					kms[i] = new MyKeyManager((X509KeyManager) kms[i], DigitalCash.getMyName());
					//System.out.println("kms["+i+"] = " + kms[i]);
				}
			}
			sslc.init(kms, tmf.getTrustManagers(), SecureRandom.getInstance("SHA1PRNG"));
		} catch (Exception e) {
			System.err.println("Unable to load keystore");
			System.exit(1);
		}
	}

	public void updateMoneyOrderList(){
		ArrayList<MoneyOrder> myVMOs = me.validMoneyOrders();
		int num = myVMOs.size() - 1;
		moNames.add(num + ": Money Order $" + myVMOs.get(num).amount());
		if(((String)moList.getItemAt(0)).equals("none")) moList.removeItemAt(0);
		moList.insertItemAt(moNames.get(num), num);
	}
}
