// Digital Cash
// Steven Kappes / Justin Fiedler

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.border.*;
import java.io.*;
import java.util.*;
import java.math.*;
import java.net.*;
import javax.net.ssl.*;
import java.security.cert.*;
import java.security.cert.Certificate;
import java.security.*;
import javax.crypto.*;
import org.w3c.dom.*;
import org.w3c.dom.Document;

public class MerchantGUI implements ActionListener {
	static final public  String gamsdir= "./"; 
	static final public String fileName = "merchant.txt";

	static private org.apache.xerces.dom.DOMImplementationImpl DOM = new org.apache.xerces.dom.DOMImplementationImpl();
	static private DocumentType docType = DOM.createDocumentType("message", "http://localhost/message.dtd", "message.dtd");

	static private Receiver rr;
	static private XMLSocketThread xmlSocket;

	final int MERCH_PORT = 3347;

	static Merchant me;
	
	// Components
	String		listen, balance, merchant, mo;
	JLabel		listenLabel, balanceLabel, moLabel, merchantLabel;
	JLabel emptyLabel = new JLabel(" ");
	JButton		listenButton, getBalanceButton, sendToBankButton;
	JTextField	listenTF, balanceTF, merchantTF;
	JComboBox		moList;

	ArrayList<String> moNames = new ArrayList<String>();

	final static String LOOKANDFEEL = null;

	public Component createComponents() {
		//CREATE COMPONENTS
		// Listen 
		listenLabel = new JLabel("");
	    	listenTF = new JTextField(20);
		listenTF.setText("Not Listening");
		listenTF.setEditable(false);
	    	listenTF.addActionListener(this);
		
		// Listen
		listenButton = new JButton("Listen for Customers");
		listenButton.setActionCommand("listen");
		listenButton.addActionListener(this);

		// Balance
		balanceLabel = new JLabel("Balance");
		balanceTF = new JTextField(20);
		balanceTF.setEditable(false);
		balanceTF.addActionListener(this);

		// Get Balance
		getBalanceButton = new JButton("Get Balance");
		getBalanceButton.setActionCommand("getbalance");
		getBalanceButton.addActionListener(this);
		
		// MO List
		moLabel = new JLabel("Money Orders");	
		String[] entries = {"none"};
		moList = new JComboBox(entries);
		moList.setSelectedIndex(0);
		moList.addActionListener(this);

		// Merchant
		//merchantLabel = new JLabel("Merchant");
	    	//merchantTF = new JTextField(20);
	    	//merchantTF.addActionListener(this);
		
		// Send MO To Merchant
		sendToBankButton = new JButton("Send");
		sendToBankButton.setActionCommand("sendtobank");
		sendToBankButton.addActionListener(this);

		//ADD TO PANE
		JPanel pane = new JPanel(new GridLayout(0, 3));
		pane.add(listenLabel);
		pane.add(emptyLabel);
		//pane.add(listenTF);
		pane.add(listenButton);
		
		pane.add(balanceLabel);
		pane.add(balanceTF);
		pane.add(getBalanceButton);

		pane.add(moLabel);
		pane.add(moList);

		//pane.add(baLabel);
		//pane.add(merchantTF);
		pane.add(sendToBankButton);

		pane.setBorder(BorderFactory.createEmptyBorder( 10, 10, 10, 10)); //(top, left, bottom, right)
		return pane;	
	}
	
	public void actionPerformed(ActionEvent e) {
		if("listen".equals(e.getActionCommand())) { //LISTEN FOR CUSTOMERS
			listenTF.setText("Listening");
			this.getMoney();
			listenTF.setText("Not listening");
		} else if("getbalance".equals(e.getActionCommand())) { //GET BALANCE
			this.requestBalance();
		} else if("sendtobank".equals(e.getActionCommand())) { //SEND TO BANK
			this.deposit();
		}
	}

	public static void setComObjects(Receiver r, XMLSocketThread xmlSoc)
	{
		rr = r;
		xmlSocket = xmlSoc;
	}

	public static void createAndShowGUI() {
		//Set the look and feel.
		try{
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		} catch(Exception e){ }

		//Make sure we have nice window decorations.
		JFrame.setDefaultLookAndFeelDecorated(true);

		//Create and set up the window.
		JFrame frame = new JFrame("Merchant GUI");
		frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);

		MerchantGUI app = new MerchantGUI();
		Component contents = app.createComponents();
		frame.getContentPane().add(contents, BorderLayout.CENTER);

		//Display the window.
		frame.pack();
		frame.setVisible(true);
	}

	public MerchantGUI()
	{
		me = new Merchant();
	}

	public static void run() {
		//Schedule a job for the event-dispatching thread:
		//creating and showing this application's GUI.
		javax.swing.SwingUtilities.invokeLater(new Runnable() { public void run() { createAndShowGUI(); } });
	}

	public void requestBalance()
	{		
		try {
			String customerID = DigitalCash.getMyIdentity().toString();
			String password = DigitalCash.MyPassword().toString();

			// Create the document
			Document doc = XMLHelper.sendBalance(customerID, password, -1);
			xmlSocket.send(doc);

			// Get the response
			doc = rr.get();
			doc.normalize();
			XMLHelper.stripText(doc);

			if (!XMLHelper.isValidMessage(doc, 0)) return;

			String amt = XMLHelper.getItem(doc, "balance", "0");
			if (amt != null)
				balanceTF.setText("$" + amt);
		} catch (Exception e) {
			System.err.println("Unable to get balance");
			System.exit(1);
		}
	}

	public void getMoney()
	{
		// Wait for a customer to connect
		try {
			SSLServerSocket ss;
			ss = (SSLServerSocket) SSLServerSocketFactory.getDefault().createServerSocket(MERCH_PORT);
			ss.setEnabledCipherSuites(ss.getSupportedCipherSuites());
			ss.setWantClientAuth(true);

			Socket sock = ss.accept();
			SSLSession sess = ((SSLSocket) sock).getSession();
			try {
				Certificate[] certs = sess.getPeerCertificates();
				System.out.println("Customer authenticated");
			} catch (SSLPeerUnverifiedException ex) {
				System.out.println("Customer NOT authenticated");
				return;
			}

			XMLSocketThread xml = new XMLSocketThread(sock);
			Receiver r2 = new Receiver(xml);
			xml.setDOMProcessor(r2);
			xml.start();

			Document doc;

			// Wait for the money order
			doc = r2.get();
			doc.normalize();
			XMLHelper.stripText(doc);

			if (!XMLHelper.isValidMessage(doc, 5)) return;

			MoneyOrder newMoney = new MoneyOrder(new ByteArray(XMLHelper.getItem(doc, "SignedHash", "0")),
					new ByteArray(XMLHelper.getItem(doc, "MoneyOrder", "0")));

			ArrayList<Integer> unlockSecs = me.receivedMoneyOrder(newMoney);
			if (unlockSecs == null) {
				xml.send(XMLHelper.sendMessage(me.ERROR));
				sess.invalidate();
				ss.close();
				return;
			}

			doc = XMLHelper.requestIdentityKeys(newMoney.uniqueString(), unlockSecs);
			xml.send(doc);

			// Wait for the response
			doc = r2.get();
			doc.normalize();
			XMLHelper.stripText(doc);

			if (!XMLHelper.isValidMessage(doc, 7)) {
				sess.invalidate();
				ss.close();
				return;
			}

			ArrayList<SecretKey> identKeys = new ArrayList<SecretKey>();
			for (int i = 0; i < DigitalCash.N; i++) {
				identKeys.add(DigitalCash.getSecretKey(XMLHelper.getItem(doc, "IdentKey", String.valueOf(i))));
			}

			if (!me.receivedIdentKeys(identKeys)) {
				xml.send(XMLHelper.sendMessage(me.ERROR));
				sess.invalidate();
				ss.close();
				return;
			}

			xml.send(XMLHelper.sendMessage("Money order accepted"));
			this.updateMoneyOrderList();
			sess.invalidate();
			ss.close();
		} catch (Exception e) {
			System.err.println("Unable to receive money from a customer: " + e.getMessage());
		}
	}

	public void deposit()
	{
		int moIndex = moList.getSelectedIndex();
		if ((moIndex >= 0) && (moIndex < me.validMoneyOrders().size())) {
			try {
				Document doc;

				// Send all the data to the bank
				MoneyOrder mo = me.validMoneyOrders().get(moIndex);
				ArrayList<Integer> secrets = mo.unlockedIdentitySecrets();
				ArrayList<SecretKey> identKeys = new ArrayList<SecretKey>();

				for (int i = 0; i < DigitalCash.N; i++) {
					if (secrets.get(i).intValue() == 0)
						identKeys.add(mo.encryptKeysLeft().get(i));
					else
						identKeys.add(mo.encryptKeysRight().get(i));				
				}
				doc = XMLHelper.cashMoneyOrder(DigitalCash.getMyIdentity().toString(), DigitalCash.MyPassword().toString(),
						mo.intString(), mo.getSignedHash(), secrets, identKeys);
				xmlSocket.send(doc);

				// Wait to see if the money was accepted
				doc = rr.get();
				doc.normalize();
				XMLHelper.stripText(doc);
				XMLHelper.isValidMessage(doc, -1); // Print the message
			} catch (Exception e) {
				System.err.println("Unable to deposit money order: " + e.getMessage());
				return;
			}
		} else {
			System.err.println("Select a money order");
		}
	}

	public void updateMoneyOrderList(){
		ArrayList<MoneyOrder> myVMOs = me.validMoneyOrders();
		int num = myVMOs.size() - 1;
		moNames.add(num + ": Money Order $" + myVMOs.get(num).amount());
		if(((String)moList.getItemAt(0)).equals("none")) moList.removeItemAt(0);
		moList.insertItemAt(moNames.get(num), num);
	}
}
