import java.util.*;
import java.math.*;
import java.security.*;
import javax.crypto.*;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.SecureRandom;
import java.security.interfaces.RSAPrivateKey;
import java.security.interfaces.RSAPublicKey;

public class main
{
	// ** Testing
	public static void main(String[] args)
	{	
		try {
			DigitalCash.Initialize();
			DigitalCash.SetMyIdentity("client");
			ByteArray custID = DigitalCash.getMyIdentity();

			Bank theBank = new Bank();
			Customer theCustomer = new Customer();
			ArrayList<ByteArray> encryptedMOs = theCustomer.generateEncryptedMoneyOrders(1000);
			ArrayList<ByteArray> blindHashes = theCustomer.getBlindedHashes();

			// Get the money order the bank will sign
			int moIndex = theBank.receivedMoneyOrders(DigitalCash.getMyIdentity(), encryptedMOs, blindHashes);
			System.out.println("Requested money order " + moIndex);

			theCustomer.setFutureOrderIndex(moIndex);

			ArrayList<SecretKey> keys = theCustomer.getAllEncryptionKeys(moIndex);
			ArrayList<BigInteger> blinds = theCustomer.getAllBlinds(moIndex);

			// Send the bank the keys
			if (!theBank.receivedEncryptionKeys(custID, keys))
				System.out.println("Encryption keys failed");

			// Send the bank the blinds
			if (!theBank.receivedUnblindFactors(custID, blinds))
				System.out.println("Blinds failed");

			// Sent the bank the identity keys
			for (int i = 0; i < DigitalCash.N; i++) {
				if (i != moIndex) {
					theBank.receivedKeys(DigitalCash.getMyIdentity(), i, theCustomer.getIdentityKeys(i, true), 
						theCustomer.getIdentityKeys(i, false));
				}
			}

			// Get the signed money order
			ByteArray signedMO = theBank.getSignedMoneyOrder(DigitalCash.getMyIdentity());

			if (signedMO != null) {
				System.out.println("Valid order? " + theCustomer.receivedSignedOrder(signedMO));
			} else {
				System.out.println("Not valid order");
				System.exit(1);
			}

			MoneyOrder validMO = theCustomer.validMoneyOrders().get(0);
			MoneyOrder toMerch = new MoneyOrder(validMO.getSignedHash(), validMO.intString());

			Merchant merch = new Merchant();
			ByteArray merchID = CryptoPrimitives.UniqueString(DigitalCash.UNIQUE_STR_SIZE);

			// Get the identity keys to return
			ArrayList<Integer> unlockSecs = merch.receivedMoneyOrder(toMerch);
			if (unlockSecs == null) {
				System.err.println("NULL");
				System.exit(1);
			}

			// Get the identity keys
			ArrayList<SecretKey> identKeys = theCustomer.getIdentKeys(toMerch.uniqueString(), unlockSecs);
			if (identKeys == null) {
				System.err.println("NULL");
				System.exit(1);
			}

			// Process the identity keys
			if (!merch.receivedIdentKeys(identKeys)) {
				System.out.println("Bad keys from the customer");
				System.exit(1);
			}

			// The bank will deposit the money order
			if (!theBank.depositMoneyOrder(merchID, toMerch)) {
				System.out.println("Unable to deposit");
				System.exit(1);
			}

			System.out.println("Deposited");

			if (!theBank.depositMoneyOrder(merchID, toMerch)) {
				System.out.println("Unable to deposit");
				System.exit(1);
			}

			System.out.println("Done");

		} catch (Exception e) {
			System.err.println("Exception: " + e.getMessage());
			e.printStackTrace();
			System.exit(1);
		}

	}
}
