// Digital Cash
// Steven Kappes / Justin Fiedler

import java.util.*;
import java.math.*;
import java.security.*;
import javax.crypto.*;

public class Merchant
{
	private int balance;

	private MoneyOrder pendingOrder;
	public String ERROR = "";
	// Digital cash
	private ArrayList<MoneyOrder> validMoneyOrders;

	public Merchant()
	{
		validMoneyOrders = new ArrayList<MoneyOrder>();
	}

	// Received a money order from a customer, request identity encryption keys
	public ArrayList<Integer> receivedMoneyOrder(MoneyOrder mo)
	{
		ArrayList<Integer> b = new ArrayList<Integer>();

		if (pendingOrder != null) {
			ERROR = "Currently in the process of receiving a money order";
			return null;
		}

		// Ensure this money has not already been spend
		for (int i = 0; i < validMoneyOrders.size(); i++) {
			if (mo.uniqueString().equals(validMoneyOrders.get(i).uniqueString())) {
				ERROR = "Invalid Money Order: This money has already been spent";
				return null;
			}
		}

		if (mo.isValidOrder()) {
			pendingOrder = mo;

			for (int i = 0; i < DigitalCash.N; i++) {
				b.add(new Integer((int)(2 * Math.random())));
				pendingOrder.unlockIdentSecret(b.get(i));
			}
			return b;
		}
		ERROR = "Invalid Money Order: Signature Unrecognized";
		return null;
	}

	// Received SecretKeys for the identity secrets
	public boolean receivedIdentKeys(ArrayList<SecretKey> identKeys)
	{
		if (pendingOrder == null) return false;
		if (identKeys.size() != DigitalCash.N) return false;

		for (int i = 0; i < identKeys.size(); i++) {
			SecretKey key = identKeys.get(i);

			// Verify the ident keys
			ByteArray identSecret = ((pendingOrder.unlockedIdentitySecrets().get(i).intValue() == 0) ?
				pendingOrder.identityStringsLeft().get(i) : pendingOrder.identityStringsRight().get(i));

			identSecret = CryptoPrimitives.AESdecrypt(identSecret, key);
			if (CryptoPrimitives.VerifyHashString(identSecret)) {
				pendingOrder.addSecretIdentKey(key);
			} else {
				ERROR = "Identity String Verification Failed";
				return false;
			}
		}

		validMoneyOrders.add(pendingOrder);
		pendingOrder = null;
		return true;
	}

	// Add a money order signed by the bank
	public void orderDeposited(int index)
	{
		balance += validMoneyOrders.get(index).amount();
	}

	// Digital cash the merchant currently has
	public ArrayList<MoneyOrder> validMoneyOrders()
	{
		return validMoneyOrders;
	}
	
	public int getBalance()
	{
		return balance;
	}
}
