import java.io.*;
import java.net.*;
import java.util.*;

import javax.net.ssl.*;
import java.security.cert.*;
import java.security.cert.Certificate;
import java.security.interfaces.*;
import java.security.*;
import org.w3c.dom.Document;

public class RunCustomer {
	static BufferedReader stdin = new BufferedReader(new InputStreamReader(System.in));
	final int PORT = 3346;	// server port to connect to

	public static void main(String[] args)  throws Exception {
		if (args.length != 2) {
			System.err.println("Please supply two command line arguments: <customer_name> <password>");
			System.exit(1);
		}
		DigitalCash.SetMyIdentity(args[0]);
		DigitalCash.getMyIdentity(); // Exit if the identity is too long
		DigitalCash.SetPassword(args[1]);
		new RunCustomer().go();
	}

	SSLContext sslc;

	static class MyKeyManager implements X509KeyManager {
		X509KeyManager x509mgr;
		String alias;

		MyKeyManager(X509KeyManager x509mgr, String alias) {
			this.x509mgr = x509mgr;
			this.alias = alias;
			String[] aliases = x509mgr.getClientAliases("RSA", null);
			for (int i=0; i<aliases.length; ++i) {
			System.out.println("found alias: " + aliases[i]);
			}
		}
		// force desired alias
		public String chooseClientAlias(String[] keytype, Principal[] issuers, Socket socket) {
			return alias;
		}
		public String chooseServerAlias(String keyType, Principal[] issuers, Socket socket) {
			return alias;
			//return x509mgr.chooseServerAlias(keyType, issuers, socket) 
		}
		public X509Certificate[] getCertificateChain(String alias) {
			return x509mgr.getCertificateChain(alias);
		}
		public PrivateKey getPrivateKey(String alias) {
			return x509mgr.getPrivateKey(alias);
		}
		public String[] getClientAliases(String keyType, Principal[] issuers) {
			return new String[] {alias};
		}
		public String[] getServerAliases(String keyType, Principal[] issuers) {
			return new String[] {alias};
		}
	}

	
	void initKeyStore() throws Exception {
		//Security.addProvider(new com.sun.net.ssl.internal.ssl.Provider());
		sslc = SSLContext.getInstance("SSL");
		KeyStore ks = KeyStore.getInstance("JKS");
		ks.load(new FileInputStream("customerstore"), "foofoo".toCharArray());
		KeyManagerFactory kmf = KeyManagerFactory.getInstance("SunX509");
		kmf.init(ks, "foofoo".toCharArray());

		KeyStore ks2 = KeyStore.getInstance("JKS");
		ks2.load(new FileInputStream("customertrust"), "barbar".toCharArray());
		TrustManagerFactory tmf = TrustManagerFactory.getInstance("SunX509");
		tmf.init(ks2);

		KeyManager[] kms = kmf.getKeyManagers();
		for (int i=0; i<kms.length; ++i) {
			if (kms[i] instanceof X509KeyManager) {
				kms[i] = new MyKeyManager((X509KeyManager) kms[i], DigitalCash.getMyName());
				System.out.println("kms["+i+"] = " + kms[i]);
			}
		}
		sslc.init(kms, tmf.getTrustManagers(),
			  SecureRandom.getInstance("SHA1PRNG"));
		//SSLSocketFactory sslsf = sslc.getSocketFactory();
		//SSLSocket ssls = (SSLSocket)sslsf.createSocket(args[0], 65535);
	}

	void go() {
		try {
			initKeyStore();
		} catch (Exception ex) {
			ex.printStackTrace();
			return;
		}

		try {
			//SSLSocket sock = (SSLSocket) SSLSocketFactory.getDefault().createSocket(InetAddress.getByName("localhost"), PORT);
			SSLSocket sock = (SSLSocket) sslc.getSocketFactory().createSocket(InetAddress.getByName("localhost"), PORT);

			sock.setEnabledCipherSuites(sock.getSupportedCipherSuites());
			sock.setNeedClientAuth(true);
			// print out some interesting SSL parameters
			SSLSession sess = sock.getSession();
			System.out.println("-- SSL Session info --");
			System.out.println("  CipherSuite = " + sess.getCipherSuite());
			System.out.println("  Protocol = " + sess.getProtocol());
			System.out.println("  PeerHost = " + sess.getPeerHost());
			try {
				Certificate[] certs = sess.getPeerCertificates();
				for (int i=0; i<certs.length; ++i) {
					System.out.println("  PeerCert = " + certs[i]);
					RSAPublicKey key = (RSAPublicKey)(certs[i].getPublicKey());
					DigitalCash.SetBankKey(key, null);
				}
			} catch (SSLPeerUnverifiedException ex) {
				System.out.println("Server NOT authenticated");
				System.exit(1);
			}
			XMLSocketThread xml = new XMLSocketThread(sock);
			Receiver rr = new Receiver(xml);
			xml.setDOMProcessor(rr);
			xml.start();

			// main loop, asks for an expression, parses it, sends it
			// to the server, and prints the response


			CustomerGUI.setComObjects(rr, xml);
			CustomerGUI.run();


			/*
			while (true) {
			System.out.println();
			System.out.println("Enter an arithmetic expression");
			System.out.print("> ");
			String line = stdin.readLine();
			line = line.trim();
			if (line.length() == 0) continue;

			Document d = new Parser().eval(line);

			System.out.println();
			System.out.println("sending to server");
			XML.format(System.out, d);

			xml.send(d);	// send it

			d = rr.get();	// wait for response
			System.out.println();
			System.out.println("received from server");
			XML.format(System.out, d);
			}
			*/

		} catch (Exception ex) {
			System.out.println("Shutting down because of exception");
			ex.printStackTrace(System.out);
		}
	}
}
