// Digital Cash
// Steven Kappes / Justin Fiedler

import java.util.*;
import java.math.*;
import java.security.interfaces.*;
import java.io.*;
import java.security.*;
import javax.crypto.*;
import javax.crypto.spec.*;
import sun.misc.*;

public class DigitalCash
{
	public static final int N = 20;
	//public static final char DELIM5 = (char)5;
	//public static final char DELIM6 = (char)6;
	public static final String ACCOUNT_PATH = "./AccountData.txt";
	private static String myID;

	// Byte sizes of the data
	public static final int UNIQUE_STR_SIZE = 16;
	public static final int UNIQUE_ENCRYPT_SIZE = 16;
	public static final int IDENTITY_STR_SIZE = 12;
	public static final int IDENTITY_ENCRYPT_SIZE = 32; // Size when encrypted with hash appended
	public static final int HASH_SIZE = CryptoPrimitives.GetHashSize();
	public static final int MO_SIZE = UNIQUE_ENCRYPT_SIZE + (2 * N * IDENTITY_ENCRYPT_SIZE) + 4 + 12; // Add 12 so this is a multiple of 16

	private static RSAPublicKey bankPubKey = null;
	private static RSAPrivateKey bankPrivKey = null;

	private static String myPass = "";
	private static Hashtable<BigInteger, String> passwords = new Hashtable<BigInteger, String>();

	public static SecretKey NULL_KEY = CryptoPrimitives.generateAESKey(); // Used as a placeholder in money orders

	public static void SetMyIdentity(String id)
	{
		myID = id;
	}

	public static void SetBankKey(RSAPublicKey pubKey, RSAPrivateKey privKey) // Customer sends null for the private key
	{
		bankPubKey = pubKey;
		bankPrivKey = privKey;
		//System.out.println("Bank public key mod: " + bankPubKey.getModulus());
		//System.out.println("Bank public key exp: " + bankPubKey.getPublicExponent());
	}

	public static void Initialize()
	{
		//generate keys for testing
		try {
		KeyPairGenerator keyGen = KeyPairGenerator.getInstance("RSA");
		keyGen.initialize(512);
		KeyPair keypair = keyGen.genKeyPair();
		bankPrivKey = (RSAPrivateKey)keypair.getPrivate();
		bankPubKey = (RSAPublicKey) keypair.getPublic();
		} catch (Exception e) {}
	}

	// Only called by the bank
	public static RSAPrivateKey getMyPrivateKey()
	{
		return bankPrivKey;
	}

	// Get the bank's public key
	public static RSAPublicKey getBankKey()
	{
		return bankPubKey;
	}

	public static boolean VerifyPassword(ByteArray custID, String pass)
	{
		if (passwords.containsKey(custID.toInteger())) {
			String p = passwords.get(custID.toInteger());
			return p.equals(pass);
		}

		// Add the password
		passwords.put(custID.toInteger(), pass);
		return true;
	}

	public static void SetPassword(String p)
	{
		myPass = p;
	}

	public static String MyPassword()
	{
		return myPass;
	}

	public static void LoadBankAccountData(Hashtable<ByteArray, Integer> balances)
	{
		try {
			MyFileReader reader = new MyFileReader(ACCOUNT_PATH, ", ");
			int numAccounts;

			balances.clear();
			
			//get size
			numAccounts = Integer.parseInt(reader.readLine());

			//read in accounts
			for (int i = 0; i < numAccounts; i++) {
				reader.readLine();
				ByteArray custID = new ByteArray((new BigInteger(reader.nextToken())).toByteArray());
				int amount = Integer.parseInt(reader.nextToken());

				// Add to the dictionary
				balances.put(custID, new Integer(amount));
			}
		} catch (Exception e) {			
			// No account data
			System.err.println("Exception: " + e.getMessage());
			e.printStackTrace();
			System.exit(1);
			balances.clear();
		}
	}

	public static void SaveBankAccountData(Hashtable<ByteArray, Integer> balances)
	{	System.out.println("Saving to file " + ACCOUNT_PATH);
    		File outFile;
    		PrintWriter pw = null;

    		outFile = new File(ACCOUNT_PATH);
    		System.out.println(ACCOUNT_PATH);
    		if(!outFile.exists()) {
    			try {
    				outFile.createNewFile();
    			} catch (Exception e) {}
    		}
    		try {
    			pw = new PrintWriter(new BufferedWriter(new FileWriter(outFile)));
    		} catch (IOException e) { }
    	
    		//print size
    		pw.println(balances.size());	//System.out.println(balances.size());
    		for(Enumeration e = balances.keys(); e.hasMoreElements();){
			ByteArray custId = (ByteArray)e.nextElement();	//System.out.println(custId);
			int bal = (int)balances.get(custId);	//System.out.println(bal);
			pw.printf("%s, %d\n", custId, bal);
		}
    		pw.close(); 		
	}

	public static ByteArray getMyIdentity()
	{
		byte [] identString = new byte[IDENTITY_STR_SIZE];
		byte [] myName = myID.getBytes();

		// Fill with spaces
		for (int i = 0; i < identString.length; i++)
			identString[i] = (byte)32;

		if (myName.length > IDENTITY_STR_SIZE) {
			System.err.println("Supplied identity is too long");
			System.exit(1);
		}

		// Copy the name
		DigitalCash.CopyAllBytes(identString, myName, 0);

		return new ByteArray(identString);
	}

	public static String getMyName()
	{
		return myID;
	}

	public static SecretKey getSecretKey(String b)
	{
		try {
			byte [] key = StringToBytes(b);
			SecretKeySpec spec = new SecretKeySpec(key, "AES");
			return (SecretKey)spec;
		} catch (Exception e) {
			System.err.println("DigitalCash.getSecretKey: Unable to generate key");
			return null;
		}
	}
	
	public static String getIntFromKey(SecretKey k)
	{
		String result = BytesToString(k.getEncoded());
		return result;
	}

	public static byte [] intToByteArray(int value)
	{
		return new byte [] {(byte)(value >>> 24), (byte)(value >> 16 & 0xff), (byte)(value >> 8 & 0xff), (byte)(value & 0xff)};
	}

	public static int byteArrayToInt(byte[] b, int offset)
	{
        int value = 0;

        for (int i = 0; i < 4; i++) {
            int shift = (4 - 1 - i) * 8;
            value += (b[i + offset] & 0x000000FF) << shift;
        }
        return value;
	}

	private static String delim = ",";
	public static String BytesToString(byte [] k)
	{
		String result = new String();

		for (int i = 0; i < k.length; i++) {
			int val = (int)k[i];
			if (result.length() > 0) result = result.concat(delim);
			result = result.concat(String.valueOf(val));
		}
		return result;
	}

	public static byte [] StringToBytes(String s)
	{
		byte [] temp;
		String [] data = s.split(",");
		byte [] k = new byte[data.length];

		for (int i = 0; i < data.length; i++) {
			int val = Integer.parseInt(data[i]);
			k[i] = (byte)val;
		}
		return k;
	}

	// Copies all of the bytes to a specific location in copyTo
	public static void CopyAllBytes(byte [] copyTo, byte [] copyFrom, int start)
	{
		for (int i = 0; i < copyFrom.length; i++) {
			copyTo[start++] = copyFrom[i];
		}
	}

	// Copies the maximum number of bytes from copyFrom that can fit in copyTo
	public static void CopySomeBytes(byte [] copyTo, byte [] copyFrom, int start)
	{
		for (int i = 0; i < copyTo.length; i++) {
			copyTo[i] = copyFrom[start++];
		}
	}
}
