import java.io.*;
import java.net.*;
import java.util.*;
import java.math.*;

import javax.net.ssl.*;
import java.security.*;
import java.security.cert.*;
import java.security.interfaces.*;
import java.security.cert.Certificate;
import java.security.spec.*;
import javax.crypto.*;

import org.w3c.dom.*;

public class Server
{
	public static void main(String[] args)
	{
		new Server().go();
	}

	SSLServerSocket ss;

 	final int PORT = 3346;	// server port to listen on

	static private org.apache.xerces.dom.DOMImplementationImpl DOM = new org.apache.xerces.dom.DOMImplementationImpl();
	static private DocumentType docType = DOM.createDocumentType("message", "http://localhost/message.dtd", "message.dtd");

	static public Bank theBank = new Bank();

    void go() {
	try {
		// Get the bank's public/private key
		RSAPublicKey pubKey = null;
		RSAPrivateKey privKey = null;
		KeyStore ks = KeyStore.getInstance("JKS");
		ks.load(new FileInputStream("bankstore"), "foofoo".toCharArray());
		KeyManagerFactory kmf = KeyManagerFactory.getInstance("SunX509");
		kmf.init(ks, "foofoo".toCharArray());

		for (Enumeration e = ks.aliases(); e.hasMoreElements() ;) {
			String alias = (String)e.nextElement();
			RSAPrivateCrtKey keyCert = (RSAPrivateCrtKey)ks.getKey(alias, "foofoo".toCharArray());

			// Get the private key
			privKey = (RSAPrivateKey)keyCert;
			
			// Get the public key
			RSAPublicKeySpec keySpec = new RSAPublicKeySpec(keyCert.getModulus(), keyCert.getPublicExponent());
			pubKey = (RSAPublicKey)KeyFactory.getInstance("RSA").generatePublic(keySpec);
		}

		DigitalCash.SetBankKey(pubKey, privKey);

	    ss = (SSLServerSocket) SSLServerSocketFactory.getDefault().createServerSocket(PORT);
	    ss.setEnabledCipherSuites(ss.getSupportedCipherSuites());
	    ss.setWantClientAuth(true);

	    // main server loop
	    // waits for a connection, creates a thread to handle the connection
	    // and goes back to waiting
	    while (true) {
		Socket sock = ss.accept();
		System.out.println("Accepted connection: " + sock);
		SSLSession sess = ((SSLSocket) sock).getSession();
		System.out.println("  SSLConnection: " + sess);
		System.out.println("  PeerHost = " + sess.getPeerHost());
		try {

		    Certificate[] certs = sess.getPeerCertificates();
		    for (int i=0; i<certs.length; ++i) {
			System.out.println("  PeerCert = " + certs[i]);
		    }
		} catch (SSLPeerUnverifiedException ex) {
		    System.out.println("Client NOT authenticated");
		}

		XMLSocketThread xml = new XMLSocketThread(sock);
		xml.setDOMProcessor(new BankReceiver(xml));
		xml.start();
	    }
	} catch (Exception ex) {
	    System.out.println("Shutting down because of exception");
	    ex.printStackTrace(System.out);
	}
    }

    /**
     * This class deals with requests from a client.  When a request is
     * received with an arithmetic expression, the value of the expression
     * is calculated and returned to the client.
     */
    class BankReceiver implements DOMProcessor {
		XMLSocketThread xml;

		BankReceiver(XMLSocketThread xml) {
			this.xml = xml;
		}

		public void process(Document doc) {
			try {
				doc.normalize();
				XMLHelper.stripText(doc);

				Document ans = eval(doc);

				if (ans != null) {
					//System.out.println();
					//System.out.println("Send back");
					//XML.format(System.out, ans);
					xml.send(ans);
				} else {
					System.out.println("Null response");
				}
			} catch (IOException ex) {
				ex.printStackTrace();
			}
		}
    }

	static Document eval(Document doc)
	{
		Document answer = null;
		int msgNum, moIndex;
		String Id, pass;
		ByteArray custID;
		MoneyOrder mo;

		ArrayList<ByteArray> encryptedMOs;
		ArrayList<ByteArray> blindHashes;
		ArrayList<SecretKey> keys;
		ArrayList<BigInteger> blinds;

		Id = XMLHelper.getCustomerID(doc);
		pass = XMLHelper.getPassword(doc);
		msgNum = XMLHelper.getMessageNumber(doc);

		if (Id == null) return null;

		custID = new ByteArray(Id);

		if (!DigitalCash.VerifyPassword(custID, pass)) {
			// Return a bad password message
			System.out.println("Bad password for " + new String(custID.toByteArray()));
			return XMLHelper.badPassword();
		}
		
		if (theBank.isBanned(custID)) {
			// The user is banned
			System.out.println(new String(custID.toByteArray()) + " is banned");
			return XMLHelper.banned();
		}

		System.out.println("Received message " + msgNum);

		switch (msgNum) {
			case 0: // Get balance
				answer = XMLHelper.sendBalance(null, null, theBank.getBalance(custID));
				break;
			case 1: // Received money orders
				encryptedMOs = new ArrayList<ByteArray>();
				blindHashes = new ArrayList<ByteArray>();

				// Get the data
				for (int i = 0; i < DigitalCash.N; i++) {
					String val;

					val = XMLHelper.getItem(doc, "EncryptedMoneyOrder", String.valueOf(i));
					if (val == null) return XMLHelper.sendMessage("Unabled to load data from XML");
					encryptedMOs.add(new ByteArray(val));

					val = XMLHelper.getItem(doc, "BlindHash", String.valueOf(i));
					if (val == null) return XMLHelper.sendMessage("Unabled to load data from XML");
					blindHashes.add(new ByteArray(val));
				}

				moIndex = theBank.receivedMoneyOrders(custID, encryptedMOs, blindHashes);
				if (moIndex == -1) {
					return XMLHelper.sendMessage(theBank.ERROR);
				}
				answer = XMLHelper.sendMOIndex(moIndex);
				break;
			case 3:
				keys = new ArrayList<SecretKey>();
				blinds = new ArrayList<BigInteger>();
				moIndex = theBank.pendingMoneyOrder(custID);

				// Get the data
				for (int i = 0; i < DigitalCash.N - 1; i++) {
					String val;

					val = XMLHelper.getItem(doc, "MoneyOrderKey", String.valueOf(i));
					if (val == null) return null;
					keys.add(DigitalCash.getSecretKey(new String(val)));

					val = XMLHelper.getItem(doc, "Blinds", String.valueOf(i));
					if (val == null) return null;
					blinds.add(new BigInteger(val));
				}

				if (!theBank.receivedEncryptionKeys(custID, keys)) {
					return XMLHelper.sendMessage(theBank.ERROR);
				}

				if (!theBank.receivedUnblindFactors(custID, blinds)) {
					return XMLHelper.sendMessage(theBank.ERROR);
				}

				// Get the identity keys
				for (int i = 0; i < DigitalCash.N - 1; i++) {
					ArrayList<SecretKey> leftKeys = new ArrayList<SecretKey>();
					ArrayList<SecretKey> rightKeys = new ArrayList<SecretKey>();

					// Get the data
					for (int j = 0; j < DigitalCash.N; j++) {
						String val;
						SecretKey key;

						val = XMLHelper.getItem(doc, "IdentKeyLeft" + String.valueOf(i), String.valueOf(j));
						if (val == null) return XMLHelper.sendMessage("Unabled to load data from XML");
						key = DigitalCash.getSecretKey(new String(val));
						leftKeys.add(key);

						val = XMLHelper.getItem(doc, "IdentKeyRight" + String.valueOf(i), String.valueOf(j));
						if (val == null) return XMLHelper.sendMessage("Unabled to load data from XML");
						key = DigitalCash.getSecretKey(new String(val));
						rightKeys.add(key);
					}
					theBank.receivedKeys(custID, (i >= moIndex ? i+1 : i), leftKeys, rightKeys);
				}

				ByteArray signedMO = theBank.getSignedMoneyOrder(custID);
				if (signedMO == null) {
					return XMLHelper.sendMessage(theBank.ERROR);
				}

				answer = XMLHelper.sendSignedHash(signedMO);
				break;
			case 8:
				String moStr = XMLHelper.getItem(doc, "MoneyOrder", "0");
				String sHash = XMLHelper.getItem(doc, "SignedHash", "0");

				mo = new MoneyOrder(new ByteArray(sHash), new ByteArray(moStr));
				for (int i = 0; i < DigitalCash.N; i++) {
					mo.unlockIdentSecret(new Integer(XMLHelper.getItem(doc, "unlocksecret", String.valueOf(i))));
					mo.addSecretIdentKey(DigitalCash.getSecretKey(XMLHelper.getItem(doc, "IdentKey", String.valueOf(i))));
				}

				if (!theBank.depositMoneyOrder(custID, mo)) {
					answer = XMLHelper.sendMessage(theBank.ERROR);
				} else {
					answer = XMLHelper.sendMessage("Deposit accepted");
				}
				break;
			default:
				break;
		}

		return answer;
    }
}
